from lqcv.data import create_reader
import cv2

dataset, _ = create_reader('0')

base = None
for img, p, _ in dataset:
    image = img
    dataset.save(save_path='./output.mp4', image=image)
    # save_path = "output.mp4"
    # if save_path != base:
    #     base = save_path
    #     fps, w, h = 30, image.shape[1], image.shape[0]
    #     vid_writer = cv2.VideoWriter(
    #         save_path, cv2.VideoWriter_fourcc(*"mp4v"), int(fps), (w, h)
    #     )
    # vid_writer.write(image)
    cv2.imshow('p', image)
    if cv2.waitKey(1) == ord('q'):
        break



import torch
import numpy as np


def clip_coords(boxes, shape):
    """
    Args:
        boxes (torch.Tensor | np.ndarray): xyxy format, (N, 4).
        shape (tuple): (height, width).
    """
    # Clip bounding xyxy bounding boxes to image shape (height, width)
    if isinstance(boxes, torch.Tensor):  # faster individually
        boxes[:, 0].clamp_(0, shape[1])  # x1
        boxes[:, 1].clamp_(0, shape[0])  # y1
        boxes[:, 2].clamp_(0, shape[1])  # x2
        boxes[:, 3].clamp_(0, shape[0])  # y2
    else:  # np.array (faster grouped)
        boxes[:, [0, 2]] = boxes[:, [0, 2]].clip(0, shape[1])  # x1, x2
        boxes[:, [1, 3]] = boxes[:, [1, 3]].clip(0, shape[0])  # y1, y2


def scale_coords(img1_shape, coords, img0_shape, ratio_pad=None):
    """
    Args:
        img1_shape (tuple): (height, width), original shape.
        coords (torch.Tensor | np.ndarray): xyxy format, (N, 4).
        img0_shape (tuple): (height, width), target shape.
    """
    # Rescale coords (xyxy) from img1_shape to img0_shape
    if ratio_pad is None:  # calculate from img0_shape
        gain = min(
            img1_shape[0] / img0_shape[0], img1_shape[1] / img0_shape[1]
        )  # gain  = old / new
        pad = (img1_shape[1] - img0_shape[1] * gain) / 2, (
            img1_shape[0] - img0_shape[0] * gain
        ) / 2  # wh padding
    else:
        gain = ratio_pad[0][0]
        pad = ratio_pad[1]

    coords[:, [0, 2]] -= pad[0]  # x padding
    coords[:, [1, 3]] -= pad[1]  # y padding
    coords[:, :4] /= gain
    clip_coords(coords, img0_shape)
    return coords


def box_iou(box1, box2):
    # https://github.com/pytorch/vision/blob/master/torchvision/ops/boxes.py
    """
    Return intersection-over-union (Jaccard index) of boxes.
    Both sets of boxes are expected to be in (x1, y1, x2, y2) format.
    Arguments:
        box1 (Tensor[N, 4])
        box2 (Tensor[M, 4])
    Returns:
        iou (Tensor[N, M]): the NxM matrix containing the pairwise
            IoU values for every element in boxes1 and boxes2
    """

    def box_area(box):
        # box = 4xn
        return (box[2] - box[0]) * (box[3] - box[1])

    area1 = box_area(box1.T)
    area2 = box_area(box2.T)

    # inter(N,M) = (rb(N,M,2) - lt(N,M,2)).clamp(0).prod(2)
    inter = (
        (
            torch.min(box1[:, None, 2:], box2[:, 2:])
            - torch.max(box1[:, None, :2], box2[:, :2])
        )
        .clamp(0)
        .prod(2)
    )
    return inter / (
        area1[:, None] + area2 - inter
    )  # iou = inter / (area1 + area2 - inter)


def nms_numpy(boxes, scores, class_id, threshold, method=None, agnostic=False):
    """
    :param boxes: numpy(N, 4), xyxy
    :param scores: numpy(N, )
    :param class_id: numpy(N, )
    :param threshold: float
    :param method:
    :return: kept boxed index
    """
    if boxes.size == 0:
        return np.empty((0,), dtype=np.int8)
    max_wh = 4096
    if isinstance(boxes, torch.Tensor):
        boxes = boxes.cpu().numpy()
    if isinstance(scores, torch.Tensor):
        scores = scores.cpu().numpy()
    if isinstance(class_id, torch.Tensor):
        class_id = class_id.cpu().numpy()

    if boxes.ndim == 1:
        boxes = boxes[None, :]
    assert boxes.shape[1] == 4, f"expected boxes shape [N, 4], but got {boxes.shape}"
    if len(class_id.shape) == 1:
        class_id = class_id[:, None]

    assert (
        boxes.shape[0] == class_id.shape[0] == scores.shape[0]
    ), f"boxes, class_id and scores shapes must be equal"

    c = class_id * (0 if agnostic else max_wh)
    boxes = boxes + c
    x1 = boxes[:, 0].copy()
    y1 = boxes[:, 1].copy()
    x2 = boxes[:, 2].copy()
    y2 = boxes[:, 3].copy()

    s = scores
    area = (x2 - x1 + 1) * (y2 - y1 + 1)

    I = np.argsort(s)  # 从小到大排序索引
    pick = np.zeros_like(s, dtype=np.int16)
    counter = 0
    while I.size > 0:
        i = I[-1]
        pick[counter] = i
        counter += 1
        idx = I[0:-1]

        xx1 = np.maximum(x1[i], x1[idx]).copy()
        yy1 = np.maximum(y1[i], y1[idx]).copy()
        xx2 = np.minimum(x2[i], x2[idx]).copy()
        yy2 = np.minimum(y2[i], y2[idx]).copy()

        w = np.maximum(0.0, xx2 - xx1 + 1).copy()
        h = np.maximum(0.0, yy2 - yy1 + 1).copy()

        inter = w * h
        if method == "Min":
            o = inter / np.minimum(area[i], area[idx])
        else:
            o = inter / (area[i] + area[idx] - inter)
        I = I[np.where(o <= threshold)]

    pick = pick[:counter].copy()
    return pick



from copy import copy
from pathlib import Path
import matplotlib
import matplotlib.pyplot as plt

# Settings
matplotlib.rc("font", **{"size": 11})
matplotlib.use("Agg")  # for writing to files only


def butter_lowpass_filtfilt(data, cutoff=1500, fs=50000, order=5):
    from scipy.signal import butter, filtfilt

    # https://stackoverflow.com/questions/28536191/how-to-filter-smooth-with-scipy-numpy
    def butter_lowpass(cutoff, fs, order):
        nyq = 0.5 * fs
        normal_cutoff = cutoff / nyq
        return butter(order, normal_cutoff, btype="low", analog=False)

    b, a = butter_lowpass(cutoff, fs, order=order)
    return filtfilt(b, a, data)  # forward-backward filter


def plot_lr_scheduler(optimizer, scheduler, epochs=300, save_dir=""):
    # Plot LR simulating training for full epochs
    optimizer, scheduler = copy(optimizer), copy(scheduler)  # do not modify originals
    y = []
    for _ in range(epochs):
        scheduler.step()
        y.append(optimizer.param_groups[0]["lr"])
    plt.plot(y, ".-", label="LR")
    plt.xlabel("epoch")
    plt.ylabel("LR")
    plt.grid()
    plt.xlim(0, epochs)
    plt.ylim(0)
    plt.savefig(Path(save_dir) / "LR.png", dpi=200)
    plt.close()



import glob
import os.path as osp
import cv2
import time
from pathlib import Path
from threading import Thread
from ..image.utils import IMG_FORMATS
from ..video.utils import VID_FORMATS
from ..utils.checker import check_requirements
from ..utils.general import clean_str

"""
NOTE:
    `ReadVideosAndImages` and `ReadStreams` are pure data readers without preprocessing.
"""


class ReadStreams:
    """Read Streams, modified from yolov5, support multi streams reading, but support one streams saving for now."""

    def __init__(self, sources="streams.txt"):
        self.mode = "stream"

        if osp.isfile(sources):
            with open(sources, "r") as f:
                sources = [
                    x.strip() for x in f.read().strip().splitlines() if len(x.strip())
                ]
        else:
            sources = [sources]

        n = len(sources)
        self.vid_path, self.vid_writer = [None] * n, [None] * n
        self.imgs, self.fps, self.frames, self.threads = (
            [None] * n,
            [0] * n,
            [0] * n,
            [None] * n,
        )
        self.sources = [clean_str(x) for x in sources]  # clean source names for later
        for i, s in enumerate(sources):  # index, source
            # Start thread to read frames from video stream
            print(f"{i + 1}/{n}: {s}... ", end="")
            if "youtube.com/" in s or "youtu.be/" in s:  # if source is YouTube video
                check_requirements(("pafy", "youtube_dl"))
                import pafy

                s = pafy.new(s).getbest(preftype="mp4").url  # YouTube URL
            s = eval(s) if s.isnumeric() else s  # i.e. s = '0' local webcam
            cap = cv2.VideoCapture(s)
            assert cap.isOpened(), f"Failed to open {s}"
            w = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
            h = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
            self.fps[i] = (
                max(cap.get(cv2.CAP_PROP_FPS) % 100, 0) or 30.0
            )  # 30 FPS fallback
            self.frames[i] = max(int(cap.get(cv2.CAP_PROP_FRAME_COUNT)), 0) or float(
                "inf"
            )  # infinite stream fallback

            _, self.imgs[i] = cap.read()  # guarantee first frame
            self.threads[i] = Thread(
                target=self.update, args=([i, cap, s]), daemon=True
            )
            print(
                f" success ({self.frames[i]} frames {w}x{h} at {self.fps[i]:.2f} FPS)"
            )
            self.threads[i].start()
        print("")  # newline

    def update(self, i, cap, stream):
        # Read stream `i` frames in daemon thread
        n, f, read = (
            0,
            self.frames[i],
            1,
        )  # frame number, frame array, inference every 'read' frame
        while cap.isOpened() and n < f:
            n += 1
            # _, self.imgs[index] = cap.read()
            cap.grab()
            if n % read == 0:
                success, im = cap.retrieve()
                if success:
                    self.imgs[i] = im
                else:
                    print(
                        "WARNING: Video stream unresponsive, please check your IP camera connection."
                    )
                    self.imgs[i] *= 0
                    cap.open(stream)  # re-open stream if signal was lost
            time.sleep(1 / self.fps[i])  # wait time

    def __iter__(self):
        self.count = -1
        return self

    def __next__(self):
        self.count += 1
        if not all(x.is_alive() for x in self.threads) or cv2.waitKey(1) == ord(
            "q"
        ):  # q to quit
            cv2.destroyAllWindows()
            raise StopIteration

        img0 = self.imgs.copy()

        # list
        return img0, self.sources, " "

    def __len__(self):
        return len(self.sources)  # 1E12 frames = 32 streams at 30 FPS for 30 years

    def save(self, save_path, image, i=0):
        # TODO: this won't work, cause the multi thread stuff.
        if self.vid_path[i] != save_path:  # new video
            self.vid_path[i] = save_path
            fps, w, h = 30, image.shape[1], image.shape[0]
            save_path += ".mp4"
            self.vid_writer[i] = cv2.VideoWriter(
                save_path, cv2.VideoWriter_fourcc(*"mp4v"), int(self.fps[i]), (w, h)
            )
        self.vid_writer[i].write(image)


class ReadOneStream:
    """Read one Stream, modified from yolov5, support one streams reading and saving."""

    def __init__(self, source):
        self.mode = "stream"

        self.vid_path, self.vid_writer = None, None
        self.fps, self.frames = 0, 0
        self.source = clean_str(source)  # clean source names for later

        # Start thread to read frames from video stream
        print(f"{1}/{1}: {source}... ", end="")
        if (
            "youtube.com/" in source or "youtu.be/" in source
        ):  # if source is YouTube video
            check_requirements(("pafy", "youtube_dl"))
            import pafy

            source = pafy.new(source).getbest(preftype="mp4").url  # YouTube URL
        source = (
            eval(source) if source.isnumeric() else source
        )  # i.e. s = '0' local webcam
        self.cap = cv2.VideoCapture(source)
        assert self.cap.isOpened(), f"Failed to open {source}"
        w = int(self.cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        h = int(self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        self.fps = (
            max(self.cap.get(cv2.CAP_PROP_FPS) % 100, 0) or 30.0
        )  # 30 FPS fallback
        self.frames = max(int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT)), 0) or float(
            "inf"
        )  # infinite stream fallback

        print(f" success ({self.frames} frames {w}x{h} at {self.fps:.2f} FPS)")
        print("")  # newline

    def __iter__(self):
        self.count = -1
        return self

    def __next__(self):
        self.count += 1
        ret_val, img0 = self.cap.read()
        if not ret_val or cv2.waitKey(1) == ord("q"):  # q to quit
            self.count += 1
            self.cap.release()
            cv2.destroyAllWindows()
            raise StopIteration

        return img0, self.source, " "

    def __len__(self):
        return 1

    def save(self, save_path, image):
        if self.vid_path != save_path:  # new video
            self.vid_path = save_path
            fps, w, h = 30, image.shape[1], image.shape[0]
            self.vid_writer = cv2.VideoWriter(
                save_path, cv2.VideoWriter_fourcc(*"mp4v"), int(self.fps), (w, h)
            )
        self.vid_writer.write(image)


class ReadVideosAndImages:
    """Read Videos and Images, modified from yolov5"""

    def __init__(self, source: str):
        p = str(Path(source).resolve())  # os-agnostic absolute path
        if "*" in p:
            files = sorted(glob.glob(p, recursive=True))  # glob
        elif osp.isdir(p):
            files = sorted(glob.glob(osp.join(p, "*.*")))  # dir
        elif osp.isfile(p):
            files = [p]  # files
        else:
            raise Exception(f"ERROR: {p} does not exist")

        # random.shuffle(files)
        images = [x for x in files if x.split(".")[-1].lower() in IMG_FORMATS]
        videos = [x for x in files if x.split(".")[-1].lower() in VID_FORMATS]
        ni, nv = len(images), len(videos)

        self.vid_path, self.vid_writer = None, None
        self.files = images + videos
        self.nf = ni + nv  # number of files
        self.video_flag = [False] * ni + [True] * nv
        self.mode = "image"
        if any(videos):
            self.new_video(videos[0])  # new video
        else:
            self.cap = None
        assert self.nf > 0, (
            f"No images or videos found in {p}. "
            f"Supported formats are:\nimages: {IMG_FORMATS}\nvideos: {VID_FORMATS}"
        )

    def __iter__(self):
        self.count = 0
        return self

    def __next__(self):
        if self.count == self.nf:
            raise StopIteration
        path = self.files[self.count]

        if self.video_flag[self.count]:
            # Read video
            self.mode = "video"
            ret_val, img0 = self.cap.read()
            if not ret_val:
                self.count += 1
                self.cap.release()
                if self.count == self.nf:  # last video
                    raise StopIteration
                else:
                    path = self.files[self.count]
                    self.new_video(path)
                    ret_val, img0 = self.cap.read()

            self.frame += 1
            s = f"video {self.count + 1}/{self.nf} "
        else:
            # Read image
            self.count += 1
            img0 = cv2.imread(path)  # BGR
            assert img0 is not None, "Image Not Found " + path
            s = f"image {self.count}/{self.nf} "

        return img0, path, s

    def new_video(self, path):
        self.frame = 0
        self.cap = cv2.VideoCapture(path)
        self.frames = int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT))

    def save(self, save_path, image):
        if self.mode == "image":
            cv2.imwrite(save_path, image)
        else:  # 'video' or 'stream'
            if self.vid_path != save_path:  # new video
                self.vid_path = save_path
                if isinstance(self.vid_writer, cv2.VideoWriter):
                    self.vid_writer.release()  # release previous video writer
                fps = self.cap.get(cv2.CAP_PROP_FPS)
                w = int(self.cap.get(cv2.CAP_PROP_FRAME_WIDTH))
                h = int(self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
                self.vid_writer = cv2.VideoWriter(
                    save_path, cv2.VideoWriter_fourcc(*"mp4v"), fps, (w, h)
                )
            self.vid_writer.write(image)

    def __len__(self):
        return self.nf  # number of files


def create_reader(source: str):
    """This is for data(video, webcam, image, image_path) reading in inference.
    Args:
        source(str): data source, could be a video,image,dir or webcam.
    Return:
        reader(ReadVideosAndImages | ReadStreams): data reader.
        webcam(bool): the source is webcam or not.
    """
    is_file = Path(source).suffix[1:] in (IMG_FORMATS + VID_FORMATS)
    is_url = source.lower().startswith(("rtsp://", "rtmp://", "http://", "https://"))
    webcam = source.isnumeric() or source.endswith(".txt") or (is_url and not is_file)
    return ReadOneStream(source) if webcam else ReadVideosAndImages(source), webcam


if __name__ == "__main__":
    test = create_reader(source="/d/九江/playphone/20211223/imgs")
    for img, p, s in test:
        print(s, p)
        cv2.imshow("p", img)
        if cv2.waitKey(1) == ord("q"):
            break



# Copyright (c) OpenMMLab. All rights reserved.
from io import BytesIO, StringIO
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, TextIO, Union

from ..utils.misc import is_list_of
from .file_client import FileClient
from .handlers import BaseFileHandler, JsonHandler, PickleHandler, YamlHandler

FileLikeObject = Union[TextIO, StringIO, BytesIO]

file_handlers = {
    'json': JsonHandler(),
    'yaml': YamlHandler(),
    'yml': YamlHandler(),
    'pickle': PickleHandler(),
    'pkl': PickleHandler()
}


def load(file: Union[str, Path, FileLikeObject],
         file_format: Optional[str] = None,
         file_client_args: Optional[Dict] = None,
         **kwargs):
    """Load data from json/yaml/pickle files.

    This method provides a unified api for loading data from serialized files.

    Note:
        In v1.3.16 and later, ``load`` supports loading data from serialized
        files those can be storaged in different backends.

    Args:
        file (str or :obj:`Path` or file-like object): Filename or a file-like
            object.
        file_format (str, optional): If not specified, the file format will be
            inferred from the file extension, otherwise use the specified one.
            Currently supported formats include "json", "yaml/yml" and
            "pickle/pkl".
        file_client_args (dict, optional): Arguments to instantiate a
            FileClient. See :class:`mmcv.fileio.FileClient` for details.
            Default: None.

    Examples:
        >>> load('/path/of/your/file')  # file is storaged in disk
        >>> load('https://path/of/your/file')  # file is storaged in Internet
        >>> load('s3://path/of/your/file')  # file is storaged in petrel

    Returns:
        The content from the file.
    """
    if isinstance(file, Path):
        file = str(file)
    if file_format is None and isinstance(file, str):
        file_format = file.split('.')[-1]
    if file_format not in file_handlers:
        raise TypeError(f'Unsupported format: {file_format}')

    handler = file_handlers[file_format]
    f: FileLikeObject
    if isinstance(file, str):
        file_client = FileClient.infer_client(file_client_args, file)
        if handler.str_like:
            with StringIO(file_client.get_text(file)) as f:
                obj = handler.load_from_fileobj(f, **kwargs)
        else:
            with BytesIO(file_client.get(file)) as f:
                obj = handler.load_from_fileobj(f, **kwargs)
    elif hasattr(file, 'read'):
        obj = handler.load_from_fileobj(file, **kwargs)
    else:
        raise TypeError('"file" must be a filepath str or a file-object')
    return obj


def dump(obj: Any,
         file: Optional[Union[str, Path, FileLikeObject]] = None,
         file_format: Optional[str] = None,
         file_client_args: Optional[Dict] = None,
         **kwargs):
    """Dump data to json/yaml/pickle strings or files.

    This method provides a unified api for dumping data as strings or to files,
    and also supports custom arguments for each file format.

    Note:
        In v1.3.16 and later, ``dump`` supports dumping data as strings or to
        files which is saved to different backends.

    Args:
        obj (any): The python object to be dumped.
        file (str or :obj:`Path` or file-like object, optional): If not
            specified, then the object is dumped to a str, otherwise to a file
            specified by the filename or file-like object.
        file_format (str, optional): Same as :func:`load`.
        file_client_args (dict, optional): Arguments to instantiate a
            FileClient. See :class:`mmcv.fileio.FileClient` for details.
            Default: None.

    Examples:
        >>> dump('hello world', '/path/of/your/file')  # disk
        >>> dump('hello world', 's3://path/of/your/file')  # ceph or petrel

    Returns:
        bool: True for success, False otherwise.
    """
    if isinstance(file, Path):
        file = str(file)
    if file_format is None:
        if isinstance(file, str):
            file_format = file.split('.')[-1]
        elif file is None:
            raise ValueError(
                'file_format must be specified since file is None')
    if file_format not in file_handlers:
        raise TypeError(f'Unsupported format: {file_format}')
    f: FileLikeObject
    handler = file_handlers[file_format]
    if file is None:
        return handler.dump_to_str(obj, **kwargs)
    elif isinstance(file, str):
        file_client = FileClient.infer_client(file_client_args, file)
        if handler.str_like:
            with StringIO() as f:
                handler.dump_to_fileobj(obj, f, **kwargs)
                file_client.put_text(f.getvalue(), file)
        else:
            with BytesIO() as f:
                handler.dump_to_fileobj(obj, f, **kwargs)
                file_client.put(f.getvalue(), file)
    elif hasattr(file, 'write'):
        handler.dump_to_fileobj(obj, file, **kwargs)
    else:
        raise TypeError('"file" must be a filename str or a file-object')


def _register_handler(handler: BaseFileHandler,
                      file_formats: Union[str, List[str]]) -> None:
    """Register a handler for some file extensions.

    Args:
        handler (:obj:`BaseFileHandler`): Handler to be registered.
        file_formats (str or list[str]): File formats to be handled by this
            handler.
    """
    if not isinstance(handler, BaseFileHandler):
        raise TypeError(
            f'handler must be a child of BaseFileHandler, not {type(handler)}')
    if isinstance(file_formats, str):
        file_formats = [file_formats]
    if not is_list_of(file_formats, str):
        raise TypeError('file_formats must be a str or a list of str')
    for ext in file_formats:
        file_handlers[ext] = handler


def register_handler(file_formats: Union[str, list], **kwargs) -> Callable:

    def wrap(cls):
        _register_handler(cls(**kwargs), file_formats)
        return cls

    return wrap



import numpy as np
import cv2
from ..bbox.convert import xyxy2xywh

def polygon2mask(img_size, polygons, color=1, downsample_ratio=1):
    """
    Args:
        img_size (tuple): The image size.
        polygons (np.ndarray): [N, M], N is the number of polygons,
            M is the number of points(Be divided by 2).
        downsample_ratio (int): downsample ratio.
    """
    mask = np.zeros(img_size, dtype=np.uint8)
    polygons = np.asarray(polygons)
    polygons = polygons.astype(np.int32)
    shape = polygons.shape
    polygons = polygons.reshape(shape[0], -1, 2)
    cv2.fillPoly(mask, polygons, color=color)
    nh, nw = (
            img_size[0] // downsample_ratio, 
            img_size[1] // downsample_ratio
            )
    mask = cv2.resize(mask, (nw, nh))
    return mask

def segment2box(segment, width=640, height=640):
    # Convert 1 segment label to 1 box label, applying inside-image constraint, i.e. (xy1, xy2, ...) to (xyxy)
    x, y = segment.T  # segment xy
    inside = (x >= 0) & (y >= 0) & (x <= width) & (y <= height)
    x, y, = (
        x[inside],
        y[inside],
    )
    return (
        np.array([x.min(), y.min(), x.max(), y.max()]) if any(x) else np.zeros((1, 4))
    )  # xyxy


def segments2boxes(segments):
    # Convert segment labels to box labels, i.e. (cls, xy1, xy2, ...) to (cls, xywh)
    boxes = []
    for s in segments:
        x, y = s.T  # segment xy
        boxes.append([x.min(), y.min(), x.max(), y.max()])  # cls, xyxy
    return xyxy2xywh(np.array(boxes))  # cls, xywh



# Copyright (c) OpenMMLab. All rights reserved.
import pickle

from .base import BaseFileHandler


class PickleHandler(BaseFileHandler):

    str_like = False

    def load_from_fileobj(self, file, **kwargs):
        return pickle.load(file, **kwargs)

    def load_from_path(self, filepath, **kwargs):
        return super().load_from_path(filepath, mode='rb', **kwargs)

    def dump_to_str(self, obj, **kwargs):
        kwargs.setdefault('protocol', 2)
        return pickle.dumps(obj, **kwargs)

    def dump_to_fileobj(self, obj, file, **kwargs):
        kwargs.setdefault('protocol', 2)
        pickle.dump(obj, file, **kwargs)

    def dump_to_path(self, obj, filepath, **kwargs):
        super().dump_to_path(obj, filepath, mode='wb', **kwargs)



# YOLOv5 🚀 by Ultralytics, GPL-3.0 license
"""
General utils
"""

import contextlib
import glob
import math
import os
import random
import re
import signal
import urllib
import collections
from itertools import repeat
from pathlib import Path

import cv2
import numpy as np
import pandas as pd
import torch


# Settings
torch.set_printoptions(linewidth=320, precision=5, profile="long")
np.set_printoptions(
    linewidth=320, formatter={"float_kind": "{:11.5g}".format}
)  # format short g, %precision=5
pd.options.display.max_columns = 10
# prevent OpenCV from multithreading (incompatible with PyTorch DataLoader)
cv2.setNumThreads(0)
os.environ["NUMEXPR_MAX_THREADS"] = str(min(os.cpu_count(), 8))  # NumExpr max threads


class Timeout(contextlib.ContextDecorator):
    # Usage: @Timeout(seconds) decorator or 'with Timeout(seconds):' context manager
    def __init__(self, seconds, *, timeout_msg="", suppress_timeout_errors=True):
        self.seconds = int(seconds)
        self.timeout_message = timeout_msg
        self.suppress = bool(suppress_timeout_errors)

    def _timeout_handler(self, signum, frame):
        raise TimeoutError(self.timeout_message)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self._timeout_handler)  # Set handler for SIGALRM
        signal.alarm(self.seconds)  # start countdown for SIGALRM to be raised

    def __exit__(self, exc_type, exc_val, exc_tb):
        signal.alarm(0)  # Cancel SIGALRM if it's scheduled
        if self.suppress and exc_type is TimeoutError:  # Suppress TimeoutError
            return True


def methods(instance):
    # Get class/instance methods
    return [
        f
        for f in dir(instance)
        if callable(getattr(instance, f)) and not f.startswith("__")
    ]


def print_args(name, opt):
    # Print argparser arguments
    opt = opt if isinstance(opt, dict) else vars(opt)
    print(colorstr(f"{name}: ") + ", ".join(f"{k}={v}" for k, v in opt.items()))


def init_seeds(seed=0):
    # Initialize random number generator (RNG) seeds https://pytorch.org/docs/stable/notes/randomness.html
    # cudnn seed 0 settings are slower and more reproducible, else faster and less reproducible
    import torch.backends.cudnn as cudnn

    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    cudnn.benchmark, cudnn.deterministic = (False, True) if seed == 0 else (True, False)


def get_latest_run(search_dir="."):
    # Return path to most recent 'last.pt' in /runs (i.e. to --resume from)
    last_list = glob.glob(f"{search_dir}/**/last*.pt", recursive=True)
    return max(last_list, key=os.path.getctime) if last_list else ""


def url2file(url):
    # Convert URL to filename, i.e. https://url.com/file.txt?auth -> file.txt
    url = str(Path(url)).replace(":/", "://")  # Pathlib turns :// -> :/
    file = Path(urllib.parse.unquote(url)).name.split("?")[
        0
    ]  # '%2F' to '/', split https://url.com/file.txt?auth
    return file


def make_divisible(x, divisor):
    # Returns x evenly divisible by divisor
    return math.ceil(x / divisor) * divisor


def clean_str(s):
    # Cleans a string by replacing special characters with underscore _
    return re.sub(pattern="[|@#!¡·$€%&()=?¿^*;:,¨´><+]", repl="_", string=s)


def colorstr(*input):
    # Colors a string https://en.wikipedia.org/wiki/ANSI_escape_code, i.e.  colorstr('blue', 'hello world')
    *args, string = (
        input if len(input) > 1 else ("blue", "bold", input[0])
    )  # color arguments, string
    colors = {
        "black": "\033[30m",  # basic colors
        "red": "\033[31m",
        "green": "\033[32m",
        "yellow": "\033[33m",
        "blue": "\033[34m",
        "magenta": "\033[35m",
        "cyan": "\033[36m",
        "white": "\033[37m",
        "bright_black": "\033[90m",  # bright colors
        "bright_red": "\033[91m",
        "bright_green": "\033[92m",
        "bright_yellow": "\033[93m",
        "bright_blue": "\033[94m",
        "bright_magenta": "\033[95m",
        "bright_cyan": "\033[96m",
        "bright_white": "\033[97m",
        "end": "\033[0m",  # misc
        "bold": "\033[1m",
        "underline": "\033[4m",
    }
    return "".join(colors[x] for x in args) + f"{string}" + colors["end"]


def increment_path(path, exist_ok=False, sep="", mkdir=False):
    # Increment file or directory path, i.e. runs/exp --> runs/exp{sep}2, runs/exp{sep}3, ... etc.
    path = Path(path)  # os-agnostic
    if path.exists() and not exist_ok:
        suffix = path.suffix
        path = path.with_suffix("")
        dirs = glob.glob(f"{path}{sep}*")  # similar paths
        matches = [re.search(rf"%s{sep}(\d+)" % path.stem, d) for d in dirs]
        i = [int(m.groups()[0]) for m in matches if m]  # indices
        n = max(i) + 1 if i else 2  # increment number
        path = Path(f"{path}{sep}{n}{suffix}")  # update path
    dir = path if path.suffix == "" else path.parent  # directory
    if not dir.exists() and mkdir:
        dir.mkdir(parents=True, exist_ok=True)  # make directory
    return path

def _ntuple(n):
    def parse(x):
        if isinstance(x, collections.abc.Iterable):
            return x
        return tuple(repeat(x, n))
    return parse

to_1tuple = _ntuple(1)
to_2tuple = _ntuple(2)



import torch
import numpy as np
from .utils import clip_coords


def xyxy2xywh(x):
    # Convert nx4 boxes from [x1, y1, x2, y2] to [x, y, w, h] where xy1=top-left, xy2=bottom-right
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[:, 0] = (x[:, 0] + x[:, 2]) / 2  # x center
    y[:, 1] = (x[:, 1] + x[:, 3]) / 2  # y center
    y[:, 2] = x[:, 2] - x[:, 0]  # width
    y[:, 3] = x[:, 3] - x[:, 1]  # height
    return y


def xywh2xyxy(x):
    # Convert nx4 boxes from [x, y, w, h] to [x1, y1, x2, y2] where xy1=top-left, xy2=bottom-right
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[:, 0] = x[:, 0] - x[:, 2] / 2  # top left x
    y[:, 1] = x[:, 1] - x[:, 3] / 2  # top left y
    y[:, 2] = x[:, 0] + x[:, 2] / 2  # bottom right x
    y[:, 3] = x[:, 1] + x[:, 3] / 2  # bottom right y
    return y


def xywh2ltwh(x):
    # Convert nx4 boxes from [x, y, w, h] to [x1, y1, w, h] where xy1=top-left
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[:, 0] = x[:, 0] - x[:, 2] / 2  # top left x
    y[:, 1] = x[:, 1] - x[:, 3] / 2  # top left y
    return y


def xyxy2ltwh(x):
    # Convert nx4 boxes from [x1, y1, x2, y2] to [x1, y1, w, h] where xy1=top-left, xy2=bottom-right
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[:, 2] = x[:, 2] - x[:, 0]  # width
    y[:, 3] = x[:, 3] - x[:, 1]  # height
    return y


def ltwh2xywh(x):
    # Convert nx4 boxes from [x1, y1, w, h] to [x, y, w, h] where xy1=top-left, xy=center
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[:, 0] = x[:, 0] + x[:, 2] / 2  # center x
    y[:, 1] = x[:, 1] + x[:, 3] / 2  # center y
    return y


def ltwh2xyxy(x):
    # Convert nx4 boxes from [x1, y1, w, h] to [x1, y1, x2, y2] where xy1=top-left, xy2=bottom-right
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[:, 2] = x[:, 2] + x[:, 0]  # width
    y[:, 3] = x[:, 3] + x[:, 1]  # height
    return y


def xywhn2xyxy(x, w=640, h=640, padw=0, padh=0):
    # Convert nx4 boxes from [x, y, w, h] normalized to [x1, y1, x2, y2] where xy1=top-left, xy2=bottom-right
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[:, 0] = w * (x[:, 0] - x[:, 2] / 2) + padw  # top left x
    y[:, 1] = h * (x[:, 1] - x[:, 3] / 2) + padh  # top left y
    y[:, 2] = w * (x[:, 0] + x[:, 2] / 2) + padw  # bottom right x
    y[:, 3] = h * (x[:, 1] + x[:, 3] / 2) + padh  # bottom right y
    return y


def xyxy2xywhn(x, w=640, h=640, clip=False, eps=0.0):
    # Convert nx4 boxes from [x1, y1, x2, y2] to [x, y, w, h] normalized where xy1=top-left, xy2=bottom-right
    if clip:
        clip_coords(x, (h - eps, w - eps))  # warning: inplace clip
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[:, 0] = ((x[:, 0] + x[:, 2]) / 2) / w  # x center
    y[:, 1] = ((x[:, 1] + x[:, 3]) / 2) / h  # y center
    y[:, 2] = (x[:, 2] - x[:, 0]) / w  # width
    y[:, 3] = (x[:, 3] - x[:, 1]) / h  # height
    return y


def xyn2xy(x, w=640, h=640, padw=0, padh=0):
    # Convert normalized segments into pixel segments, shape (n,2)
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[:, 0] = w * x[:, 0] + padw  # top left x
    y[:, 1] = h * x[:, 1] + padh  # top left y
    return y



import numpy as np
import cv2
import torch


def resample_segments(segments, n=1000):
    # Up-sample an (n,2) segment
    for i, s in enumerate(segments):
        x = np.linspace(0, len(s) - 1, n)
        xp = np.arange(len(s))
        segments[i] = (
            np.concatenate([np.interp(x, xp, s[:, i]) for i in range(2)])
            .reshape(2, -1)
            .T
        )  # segment xy
    return segments

def scale_masks(img1_shape, masks, img0_shape, ratio_pad=None):
    """
    img1_shape: model input shape, [h, w]
    img0_shape: origin pic shape, [h, w, 3]
    masks: [h, w, num]
    resize for the most time
    """
    # Rescale coords (xyxy) from img1_shape to img0_shape
    if ratio_pad is None:  # calculate from img0_shape
        gain = min(img1_shape[0] / img0_shape[0],
                   img1_shape[1] / img0_shape[1])  # gain  = old / new
        pad = (img1_shape[1] - img0_shape[1] * gain) / 2, (
            img1_shape[0] - img0_shape[0] * gain) / 2  # wh padding
    else:
        gain = ratio_pad[0][0]
        pad = ratio_pad[1]
    tl_pad = int(pad[1]), int(pad[0])  # y, x
    br_pad = int(img1_shape[0] - pad[1]), int(img1_shape[1] - pad[0])
    
    if len(masks.shape) < 2:
        raise ValueError(f'"len of masks shape" should be 2 or 3, but got {len(masks.shape)}')
    if len(masks.shape) == 2:
        masks = masks[:, :, None]
    # masks_h, masks_w, n
    masks = masks[tl_pad[0]:br_pad[0], tl_pad[1]:br_pad[1], :]
    masks = cv2.resize(masks, (img0_shape[1], img0_shape[0]))

    return masks

def mask_iou(mask1, mask2):
    """
    mask1: [N, n] m1 means number of predicted objects 
    mask2: [M, n] m2 means number of gt objects
    Note: n means image_w x image_h

    return: masks iou, [N, M]
    """
    # print(mask1.shape)
    # print(mask2.shape)
    intersection = torch.matmul(mask1, mask2.t()).clamp(0)
    area1 = torch.sum(mask1, dim=1).view(1, -1)
    area2 = torch.sum(mask2, dim=1).view(1, -1)
    union = (area1.t() + area2) - intersection

    return intersection / (union + 1e-7)



import numpy as np
import torch


class Boxes:
    """
    Args:
        boxes (np.ndarray | torch.Tensor): boxes, xyxy format.
    """

    def __init__(self, boxes) -> None:
        boxes = boxes[None, :] if boxes.ndim == 1 else boxes
        assert boxes.ndim == 2
        assert boxes.shape[1] == 4

        # (n, )
        self.x1 = boxes[:, 0]
        self.y1 = boxes[:, 1]
        self.x2 = boxes[:, 2]
        self.y2 = boxes[:, 3]
        # n, 4
        self.boxes = boxes

    def stack(self, *args, **kwargs):
        return (
            torch.stack(*args, **kwargs)
            if isinstance(self.boxes, torch.Tensor)
            else np.stack(*args, **kwargs)
        )

    def cat(self):
        return torch.cat if isinstance(self.boxes, torch.Tensor) else np.concatenate

    @property
    def lt(self):
        """left top, (n, 2)"""
        return self.stack([self.x1, self.y1], 1)

    @property
    def rt(self):
        """right top, (n, 2)"""
        return self.stack([self.x2, self.y1], 1)

    @property
    def lb(self):
        """left bottom, (n, 2)"""
        return self.stack([self.x1, self.y2], 1)

    @property
    def rb(self):
        """right bottom, (n, 2)"""
        return self.stack([self.x2, self.y2], 1)

    @property
    def center(self):
        """center, (n, 2)"""
        x_center = (self.x1 + self.x2) / 2
        y_center = (self.y1 + self.y2) / 2
        return self.stack([x_center, y_center], 1)

    @property
    def areas(self):
        areas = (self.boxes[:, 2] - self.boxes[:, 0]) * (
            self.boxes[:, 3] - self.boxes[:, 1]
        )
        return areas

    def get_vertice(self, filter=None, frame=None):
        """Get the vertice of self.boxes.
        Args:
            filter (str | List[str] | [optional]): include `lt`,
                `lb`, `rt`, `rb` or `center`, if None, return
                `lt` + `lb` + `rt` + `rb`.
            frame (ndarray): image for visualization.
        Return:
            same type with self.boxes, shape: (m, n, 2),
                n is number of boxes, m if number of vertices.
        """
        if filter is None:
            # NOTE: exclude center
            vertices = self.stack([self.lt, self.lb, self.rt, self.rb], 0)
        else:
            if isinstance(filter, str):
                filter = [filter]
            coords = []
            for f in filter:
                f in ["lt", "lb", "rt", "rb", "center"]
                if f == "lt":
                    coords.append(self.lt)
                elif f == "lb":
                    coords.append(self.lb)
                elif f == "rt":
                    coords.append(self.rt)
                elif f == "rb":
                    coords.append(self.rb)
                else:
                    coords.append(self.center)
            vertices = (
                coords[0][None, :, :] if len(coords) == 1 else self.stack(coords, 0)
            )
        if frame is not None:
            import cv2
            for coords in vertices.tolist():
                for coord in coords:
                    cv2.circle(frame, (int(coord[0]), int(coord[1])), 1, (0, 0, 255), 5)
        return vertices



# from mmcv
from time import time
from time import sleep
import torch
import functools


def time_sync():
    # pytorch-accurate time
    if torch.cuda.is_available():
        torch.cuda.synchronize()
    return time()


class TimerError(Exception):
    def __init__(self, message):
        self.message = message
        super(TimerError, self).__init__(message)


class Timer:
    """A flexible Timer class.

    :Example:

    >>> import time
    >>> import mmcv
    >>> with mmcv.Timer():
    >>>     # simulate a code block that will run for 1s
    >>>     time.sleep(1)
    1.000
    >>> with mmcv.Timer(print_tmpl='it takes {:.1f} seconds'):
    >>>     # simulate a code block that will run for 1s
    >>>     time.sleep(1)
    it takes 1.0 seconds
    >>> timer = mmcv.Timer()
    >>> time.sleep(0.5)
    >>> print(timer.since_start())
    0.500
    >>> time.sleep(0.5)
    >>> print(timer.since_last_check())
    0.500
    >>> print(timer.since_start())
    1.000
    """

    def __init__(
        self, start=True, print_tmpl=None, cuda_sync=False, round=None, unit="s"
    ):
        assert unit in ['s', 'ms']
        assert round is None or isinstance(round, int)
        self._is_running = False
        self.round = round
        self.unit = unit
        self.print_tmpl = print_tmpl if print_tmpl else "{:.3f}"
        self.time = time_sync if cuda_sync else time
        if start:
            self.start()

    @property
    def is_running(self):
        """bool: indicate whether the timer is running"""
        return self._is_running

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, type, value, traceback):
        print(self.print_tmpl.format(self.since_last_check()))
        self._is_running = False

    def start(self, reset=False):
        """Start the timer."""
        if not self._is_running:
            self._t_start = self.time()
            self._is_running = True
        self._t_start = self.time() if reset else self._t_start
        self._t_last = self.time()

    def since_start(self):
        """Total time since the timer is started.

        Returns (float): Time in seconds.
        """
        if not self._is_running:
            raise TimerError("timer is not running")
        self._t_last = self.time()
        dur = self._t_last - self._t_start
        if self.unit == 'ms':
            dur = dur * 1000
        if self.round is not None:
            dur = round(dur, self.round)
        return dur

    def since_last_check(self):
        """Time since the last checking.

        Either :func:`since_start` or :func:`since_last_check` is a checking
        operation.

        Returns (float): Time in seconds.
        """
        if not self._is_running:
            raise TimerError("timer is not running")
        dur = self.time() - self._t_last
        if self.unit == 'ms':
            dur = dur * 1000
        if self.round is not None:
            dur = round(dur, self.round)
        self._t_last = self.time()
        return dur


_g_timers = {}  # global timers


def check_time(timer_id):
    """Add check points in a single line.

    This method is suitable for running a task on a list of items. A timer will
    be registered when the method is called for the first time.

    :Example:

    >>> import time
    >>> import mmcv
    >>> for i in range(1, 6):
    >>>     # simulate a code block
    >>>     time.sleep(i)
    >>>     mmcv.check_time('task1')
    2.000
    3.000
    4.000
    5.000

    Args:
        timer_id (str): Timer identifier.
    """
    if timer_id not in _g_timers:
        _g_timers[timer_id] = Timer()
        return 0
    else:
        return _g_timers[timer_id].since_last_check()


def timer(cuda_sync=False):
    def out_func(old_func):
        time = Timer(start=False, cuda_sync=cuda_sync)

        @functools.wraps(old_func)
        def new_func(*args, **kwargs):
            time.start()
            output = old_func(*args, **kwargs)
            print(f"timer: using {time.since_start()} s")
            return output

        return new_func

    return out_func


if __name__ == "__main__":
    timer = Timer()
    sleep(0.5)
    print(timer.since_start())
    sleep(0.5)
    print(timer.since_last_check())
    print(timer.since_start())



import os
import shutil
from tqdm import tqdm
from pathlib import Path
import glob

from .misc import is_str


def is_filepath(x):
    return is_str(x) or isinstance(x, Path)


def create_folder(path="./new"):
    # Create folder
    if os.path.exists(path):
        shutil.rmtree(path)  # delete output folder
    os.makedirs(path)  # make new output folder


def flatten_recursive(path="../datasets/coco128"):
    # Flatten a recursive directory by bringing all files to top level
    new_path = Path(path + "_flat")
    create_folder(new_path)
    for file in tqdm(glob.glob(str(Path(path)) + "/**/*.*", recursive=True)):
        shutil.copyfile(file, new_path / Path(file).name)


def check_file_exist(filename, msg_tmpl='file "{}" does not exist'):
    if not os.path.isfile(filename):
        raise FileNotFoundError(msg_tmpl.format(filename))


def mkdir_or_exist(dir_name, mode=0o777):
    if dir_name == "":
        return
    dir_name = os.path.expanduser(dir_name)
    os.makedirs(dir_name, mode=mode, exist_ok=True)



# YOLOv5 🚀 by Ultralytics, GPL-3.0 license
"""
Plotting utils
"""


import cv2
import numpy as np
import torch
from PIL import Image, ImageDraw, ImageFont
from itertools import repeat

from .color import colors


class Annotator:
    def __init__(
        self,
        im,
        line_width=None,
        font_size=None,
        font=None,
    ):
        assert (
            im.data.contiguous
        ), "Image not contiguous. Apply np.ascontiguousarray(im) to Annotator() input images."
        self.pil = font is not None
        if self.pil:  # use PIL
            self.im = im if isinstance(im, Image.Image) else Image.fromarray(im)
            self.draw = ImageDraw.Draw(self.im)
            self.font = ImageFont.truetype(
                str(font),
                size=font_size or max(round(sum(self.im.size) / 2 * 0.035), 12),
            )
        else:  # use cv2
            self.im = im
        self.lw = line_width or max(round(sum(im.shape) / 2 * 0.003), 2)  # line width

    def box_label(
        self, box, label="", color=(128, 128, 128), txt_color=(255, 255, 255)
    ):
        # Add one xyxy box to image with label
        if self.pil:
            self.draw.rectangle(box, width=self.lw, outline=color)  # box
            if label:
                w, h = self.font.getsize(label)  # text width, height
                outside = box[1] - h >= 0  # label fits outside box
                self.draw.rectangle(
                    [
                        box[0],
                        box[1] - h if outside else box[1],
                        box[0] + w + 1,
                        box[1] + 1 if outside else box[1] + h + 1,
                    ],
                    fill=color,
                )
                # self.draw.text((box[0], box[1]), label, fill=txt_color, font=self.font, anchor='ls')  # for PIL>8.0
                self.draw.text(
                    (box[0], box[1] - h if outside else box[1]),
                    label,
                    fill=txt_color,
                    font=self.font,
                )
        else:  # cv2
            p1, p2 = (int(box[0]), int(box[1])), (int(box[2]), int(box[3]))
            cv2.rectangle(
                self.im, p1, p2, color, thickness=self.lw, lineType=cv2.LINE_AA
            )
            if label:
                tf = max(self.lw - 1, 1)  # font thickness
                w, h = cv2.getTextSize(label, 0, fontScale=self.lw / 3, thickness=tf)[
                    0
                ]  # text width, height
                outside = p1[1] - h - 3 >= 0  # label fits outside box
                p2 = p1[0] + w, p1[1] - h - 3 if outside else p1[1] + h + 3
                cv2.rectangle(self.im, p1, p2, color, -1, cv2.LINE_AA)  # filled
                cv2.putText(
                    self.im,
                    label,
                    (p1[0], p1[1] - 2 if outside else p1[1] + h + 2),
                    0,
                    self.lw / 3,
                    txt_color,
                    thickness=tf,
                    lineType=cv2.LINE_AA,
                )

    def rectangle(self, xy, fill=None, outline=None, width=1):
        # Add rectangle to image (PIL-only)
        self.draw.rectangle(xy, fill, outline, width)

    def text(self, xy, text, txt_color=(255, 255, 255)):
        # Add text to image (PIL-only)
        _, h = self.font.getsize(text)  # text width, height
        self.draw.text((xy[0], xy[1] - h + 1), text, fill=txt_color, font=self.font)

    def result(self):
        # Return annotated image as array
        return np.asarray(self.im)


class Visualizer(object):
    """Visualization of one model."""

    def __init__(self, names) -> None:
        super().__init__()
        self.names = names

    def draw_one_img(self, img, output, vis_conf=0.4):
        """Visualize one images.

        Args:
            imgs (numpy.ndarray): one image.
            outputs (torch.Tensor): one output, (num_boxes, classes+5)
            vis_confs (float, optional): Visualize threshold.
        Return:
            img (numpy.ndarray): Image after visualization.
        """
        if isinstance(output, list):
            output = output[0]
        if output is None or len(output) == 0:
            return img
        for (*xyxy, conf, cls) in reversed(output[:, :6]):
            if conf < vis_conf:
                continue
            label = "%s %.2f" % (self.names[int(cls)], conf)
            color = colors(int(cls))
            plot_one_box(xyxy, img, label=label, color=color, line_thickness=2)
        return img

    def draw_multi_img(self, imgs, outputs, vis_confs=0.4):
        """Visualize multi images.

        Args:
            imgs (List[numpy.array]): multi images.
            outputs (List[torch.Tensor]): multi outputs, List[num_boxes, classes+5].
            vis_confs (float | tuple[float], optional): Visualize threshold.
        Return:
            imgs (List[numpy.ndarray]): Images after visualization.
        """
        if isinstance(vis_confs, float):
            vis_confs = list(repeat(vis_confs, len(imgs)))
        assert len(imgs) == len(outputs) == len(vis_confs)
        for i, output in enumerate(outputs):  # detections per image
            self.draw_one_img(imgs[i], output, vis_confs[i])
        return imgs

    def draw_imgs(self, imgs, outputs, vis_confs=0.4):
        if isinstance(imgs, np.ndarray):
            return self.draw_one_img(imgs, outputs, vis_confs)
        else:
            return self.draw_multi_img(imgs, outputs, vis_confs)

    def __call__(self, imgs, outputs, vis_confs=0.4):
        return self.draw_imgs(imgs, outputs, vis_confs)



def plot_one_box(x, img, color=None, label=None, line_thickness=None):
    import random

    # Plots one bounding box on image img
    tl = (
        line_thickness or round(0.002 * (img.shape[0] + img.shape[1]) / 2) + 1
    )  # line/font thickness
    color = color or [random.randint(0, 255) for _ in range(3)]
    c1, c2 = (int(x[0]), int(x[1])), (int(x[2]), int(x[3]))
    cv2.rectangle(img, c1, c2, color, thickness=tl, lineType=cv2.LINE_AA)
    if label:
        tf = max(tl - 1, 1)  # font thickness
        t_size = cv2.getTextSize(label, 0, fontScale=tl / 3, thickness=tf)[0]
        c2 = c1[0] + t_size[0], c1[1] - t_size[1] - 3
        cv2.rectangle(img, c1, c2, color, -1, cv2.LINE_AA)  # filled
        cv2.putText(
            img,
            label,
            (c1[0], c1[1] - 2),
            0,
            tl / 3,
            [225, 255, 255],
            thickness=tf,
            lineType=cv2.LINE_AA,
        )


def plot_masks(img, masks, colors, alpha=0.5):
    """
    Args:
        img (tensor): img on cuda, shape: [3, h, w], range: [0, 1]
        masks (tensor): predicted masks on cuda, shape: [n, h, w]
        colors (List[List[Int]]): colors for predicted masks, [[r, g, b] * n]
    Return:
        img after draw masks, shape: [h, w, 3]

    transform colors and send img_gpu to cpu for the most time.
    """
    img_gpu = img.clone()
    num_masks = len(masks)
    # [n, 1, 1, 3]
    # faster this way to transform colors
    colors = torch.tensor(colors, device=img.device).float() / 255.0
    colors = colors[:, None, None, :]
    # [n, h, w, 1]
    masks = masks[:, :, :, None]
    masks_color = masks.repeat(1, 1, 1, 3) * colors * alpha
    inv_alph_masks = masks * (-alpha) + 1
    masks_color_summand = masks_color[0]
    if num_masks > 1:
        inv_alph_cumul = inv_alph_masks[: (num_masks - 1)].cumprod(dim=0)
        masks_color_cumul = masks_color[1:] * inv_alph_cumul
        masks_color_summand += masks_color_cumul.sum(dim=0)

    # print(inv_alph_masks.prod(dim=0).shape) # [h, w, 1]
    img_gpu = img_gpu.flip(dims=[0])  # filp channel for opencv
    img_gpu = img_gpu.permute(1, 2, 0).contiguous()
    # [h, w, 3]
    img_gpu = img_gpu * inv_alph_masks.prod(dim=0) + masks_color_summand
    return (img_gpu * 255).byte().cpu().numpy()



from PIL import Image, ExifTags


# Parameters
HELP_URL = "https://github.com/ultralytics/yolov5/wiki/Train-Custom-Data"
IMG_FORMATS = [
    "bmp",
    "jpg",
    "jpeg",
    "png",
    "tif",
    "tiff",
    "dng",
    "webp",
    "mpo",
]  # acceptable image suffixes

# Get orientation exif tag
for orientation in ExifTags.TAGS.keys():
    if ExifTags.TAGS[orientation] == "Orientation":
        break

def exif_size(img):
    # Returns exif-corrected PIL size
    s = img.size  # (width, height)
    try:
        rotation = dict(img._getexif().items())[orientation]
        if rotation == 6:  # rotation 270
            s = (s[1], s[0])
        elif rotation == 8:  # rotation 90
            s = (s[1], s[0])
    except:
        pass

    return s


def exif_transpose(image):
    """
    Transpose a PIL image accordingly if it has an EXIF Orientation tag.
    Inplace version of https://github.com/python-pillow/Pillow/blob/master/src/PIL/ImageOps.py exif_transpose()

    :param image: The image to transpose.
    :return: An image.
    """
    exif = image.getexif()
    orientation = exif.get(0x0112, 1)  # default 1
    if orientation > 1:
        method = {
            2: Image.FLIP_LEFT_RIGHT,
            3: Image.ROTATE_180,
            4: Image.FLIP_TOP_BOTTOM,
            5: Image.TRANSPOSE,
            6: Image.ROTATE_270,
            7: Image.TRANSVERSE,
            8: Image.ROTATE_90,
        }.get(orientation)
        if method is not None:
            image = image.transpose(method)
            del exif[0x0112]
            image.info["exif"] = exif.tobytes()
    return image





import cv2
import time
from tqdm import tqdm
import os
from pathlib import Path

"""
使用opencv捕获摄像头,并保存为视频
"""
"""
mode each:每个视频的图片单独放一个文件夹
mode all：所有视频的图片都放在一个文件夹
"""
mode = "all"
interval = 6  # 多少秒保存一次
count_only = False

vid_formats = [
    ".mov",
    ".avi",
    ".mp4",
    ".mpg",
    ".mpeg",
    ".m4v",
    ".wmv",
    ".mkv",
    ".ts",
    ".hmv",
    ".vdo",
]
# video_dir可以是视频，也可以是装有多个视频的文件夹
video_dir = "/d/九江/passengers/other_suit/大堂经理_datasets1"
save_dir = "/d/九江/passengers/other_suit/images1"

rename = False

tail = ""

resize_frame = False
resize_shape = (1280, 720)

if not os.path.exists(save_dir):
    os.mkdir(save_dir)

if os.path.isfile(video_dir):
    videos = [video_dir]
else:
    videos = sorted(
        [
            video
            for video in os.listdir(video_dir)
            if os.path.splitext(video)[-1].lower() in vid_formats
        ]
    )
print(videos)

nf = len(videos)

total_pic = 0
new_name = -1

for i, video in enumerate(videos):
    new_name += 1
    if mode == "each":
        if not os.path.exists(os.path.join(save_dir, video)):
            os.mkdir(os.path.join(save_dir, video))
    # print(video_dir, video)
    # print(os.path.join(video_dir, video))
    # exit()
    use_video = os.path.join(video_dir, video)
    cap = cv2.VideoCapture(use_video)
    nframes = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    # print(nframes)
    w = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    h = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    # fps读出来是inf，int(inf)会报错
    try:
        fps = int(cap.get(cv2.CAP_PROP_FPS))
    except:
        fps = 25
    if count_only:
        if nframes > 0:
            total_pic += nframes / (fps * interval if fps <= 100 else 25 * interval)
        else:
            print(f"nframes should > 0, but {use_video} got {nframes} nframes.")
        continue
    # print(fps)
    # continue
    # size=(1280,720)

    # video_name = os.path.split(video)[0]
    # print(video_name)
    video_name = Path(use_video).name.split(".")[0] if not rename else new_name

    print(video_name)
    # continue

    frame_num = -1

    while cap.isOpened():
        frame_num += 1
    # pbar = tqdm(range(nframes), total=nframes)
    # for frame_num in pbar:
        # print('frame_num:', frame_num)

        ret, frame = cap.read()
        if resize_frame:
            frame = cv2.resize(frame, resize_shape)
            w, h = resize_shape

        if not ret:
            print("==============视频播放完毕=============")
            break

        # if frame_num % (fps * interval) == 0:
        if (
            frame_num % (fps * interval if fps <= 30 else 25 * interval) == 0
        ):  # 有时候会有读取帧数为1000的情况
            # if 1:  # 有时候会有读取帧数为1000的情况
            # frame = cv2.resize(frame, size, interpolation=cv2.INTER_AREA)
            total_pic += 1
            log = "video:%g/%g frame:(%g/%g) W×H:(%g/%g) fps:%g %s (total:%g)" % (
                i + 1,
                nf,
                frame_num,
                nframes,
                w,
                h,
                fps,
                use_video,
                total_pic,
            )
            print(log)
            # pbar.desc = log
            if mode == "each":
                cv2.imwrite(
                    os.path.join(
                        os.path.join(save_dir, video),
                        f"{video_name}_{frame_num}_{tail}.jpg",
                    ),
                    frame,
                )
            else:
                # print(os.path.join(save_dir, f'{video_name}_{frame_num}_{tail}.jpg'))
                # print(save_dir)
                # print(video_name)
                cv2.imwrite(
                    os.path.join(save_dir, f"{video_name}_{frame_num}_{tail}.jpg"),
                    # f'{frame_num}.jpg'),
                    frame,
                )

    cap.release()
    cv2.destroyAllWindows()

if count_only:
    print(f"{nf} videos will generate {total_pic} pics({interval}s/pic)")



class Colors:
    # Ultralytics color palette https://ultralytics.com/
    def __init__(self):
        # hex = matplotlib.colors.TABLEAU_COLORS.values()
        hex = (
            "FF3838",
            "FF9D97",
            "FF701F",
            "FFB21D",
            "CFD231",
            "48F90A",
            "92CC17",
            "3DDB86",
            "1A9334",
            "00D4BB",
            "2C99A8",
            "00C2FF",
            "344593",
            "6473FF",
            "0018EC",
            "8438FF",
            "520085",
            "CB38FF",
            "FF95C8",
            "FF37C7",
        )
        self.palette = [self.hex2rgb("#" + c) for c in hex]
        self.n = len(self.palette)

    def __call__(self, i, bgr=False):
        c = self.palette[int(i) % self.n]
        return (c[2], c[1], c[0]) if bgr else c

    @staticmethod
    def hex2rgb(h):  # rgb order (PIL)
        return tuple(int(h[1 + i : 1 + i + 2], 16) for i in (0, 2, 4))


colors = Colors()  # create instance for 'from utils.plots import colors'



import re
import setuptools
# import torch

# torch_ver = [int(x) for x in torch.__version__.split(".")[:2]]
# assert torch_ver >= [1, 7], "Requires PyTorch >= 1.7"

with open("lqcv/__init__.py", "r") as f:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
        f.read(), re.MULTILINE
    ).group(1)


with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()


setuptools.setup(
    name="lqcv",
    version=version,
    author="Laughing-q",
    python_requires=">=3.6",
    long_description=long_description,
    classifiers=["Programming Language :: Python :: 3", "Operating System :: OS Independent"],
    # cmdclass={"build_ext": torch.utils.cpp_extension.BuildExtension},
    packages=setuptools.find_packages(),
)



from pathlib import Path
from subprocess import check_output
from PIL import ImageFont

import os
import glob
import pkg_resources as pkg
import torch
import re
import urllib
import platform

from .general import colorstr, make_divisible

def try_except(func):
    # try-except function. Usage: @try_except decorator
    def handler(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except Exception as e:
            print(e)

    return handler

def is_writeable(dir, test=False):
    # Return True if directory has write permissions, test opening a file with write permissions if test=True
    if test:  # method 1
        file = Path(dir) / "tmp.txt"
        try:
            with open(file, "w"):  # open file with write permissions
                pass
            file.unlink()  # remove file
            return True
        except IOError:
            return False
    else:  # method 2
        return os.access(dir, os.R_OK)  # possible issues on Windows


def is_docker():
    # Is environment a Docker container?
    return Path("/workspace").exists()  # or Path('/.dockerenv').exists()

def is_pip():
    # Is file in a pip package?
    return "site-packages" in Path(__file__).resolve().parts


def is_ascii(s=""):
    # Is string composed of all ASCII (no UTF) characters? (note str().isascii() introduced in python 3.7)
    s = str(s)  # convert list, tuple, None, etc. to str
    return len(s.encode().decode("ascii", "ignore")) == len(s)


def is_chinese(s="人工智能"):
    # Is string composed of any Chinese characters?
    return re.search("[\u4e00-\u9fff]", s)


def emojis(str=""):
    # Return platform-dependent emoji-safe version of string
    return (
        str.encode().decode("ascii", "ignore")
        if platform.system() == "Windows"
        else str
    )


def file_size(path):
    # Return file/dir size (MB)
    path = Path(path)
    if path.is_file():
        return path.stat().st_size / 1e6
    elif path.is_dir():
        return sum(f.stat().st_size for f in path.glob("**/*") if f.is_file()) / 1e6
    else:
        return 0.0


def check_online():
    # Check internet connectivity
    import socket

    try:
        socket.create_connection(("1.1.1.1", 443), 5)  # check host accessibility
        return True
    except OSError:
        return False


@try_except
def check_git_status():
    # Recommend 'git pull' if code is out of date
    msg = ", for updates see https://github.com/ultralytics/yolov5"
    print(colorstr("github: "), end="")
    assert Path(".git").exists(), "skipping check (not a git repository)" + msg
    assert not is_docker(), "skipping check (Docker image)" + msg
    assert check_online(), "skipping check (offline)" + msg

    cmd = "git fetch && git config --get remote.origin.url"
    url = (
        check_output(cmd, shell=True, timeout=5).decode().strip().rstrip(".git")
    )  # git fetch
    branch = (
        check_output("git rev-parse --abbrev-ref HEAD", shell=True).decode().strip()
    )  # checked out
    n = int(
        check_output(f"git rev-list {branch}..origin/master --count", shell=True)
    )  # commits behind
    if n > 0:
        s = f"⚠️ YOLOv5 is out of date by {n} commit{'s' * (n > 1)}. Use `git pull` or `git clone {url}` to update."
    else:
        s = f"up to date with {url} ✅"
    print(emojis(s))  # emoji-safe


def check_python(minimum="3.6.2"):
    # Check current python version vs. required python version
    check_version(platform.python_version(), minimum, name="Python ")


def check_version(
    current="0.0.0",
    minimum="0.0.0",
    name="version ",
    pinned=False,
    hard=False,
    verbose=False,
):
    # Check version vs. required version
    current, minimum = (pkg.parse_version(x) for x in (current, minimum))
    result = (current == minimum) if pinned else (current >= minimum)  # bool
    s = f"{name}{minimum} required by YOLOv5, but {name}{current} is currently installed"  # string
    if hard:
        assert result, s  # assert min requirements met
    if verbose and not result:
        import logging
        logging.warning(s)
    return result


@try_except
def check_requirements(
    requirements="requirements.txt", exclude=(), install=True
):
    # Check installed dependencies meet requirements (pass *.txt file or list of packages)
    prefix = colorstr("red", "bold", "requirements:")
    check_python()  # check python version
    if isinstance(requirements, (str, Path)):  # requirements.txt file
        file = Path(requirements)
        assert file.exists(), f"{prefix} {file.resolve()} not found, check failed."
        requirements = [
            f"{x.name}{x.specifier}"
            for x in pkg.parse_requirements(file.open())
            if x.name not in exclude
        ]
    else:  # list or tuple of packages
        requirements = [x for x in requirements if x not in exclude]

    n = 0  # number of packages updates
    for r in requirements:
        try:
            pkg.require(r)
        except Exception as e:  # DistributionNotFound or VersionConflict if requirements not met
            s = f"{prefix} {r} not found and is required by YOLOv5"
            if install:
                print(f"{s}, attempting auto-update...")
                try:
                    assert check_online(), f"'pip install {r}' skipped (offline)"
                    print(check_output(f"pip install '{r}'", shell=True).decode())
                    n += 1
                except Exception as e:
                    print(f"{prefix} {e}")
            else:
                print(f"{s}. Please install and rerun your command.")

    if n:  # if packages updated
        source = file.resolve() if "file" in locals() else requirements
        s = (
            f"{prefix} {n} package{'s' * (n > 1)} updated per {source}\n"
            f"{prefix} ⚠️ {colorstr('bold', 'Restart runtime or rerun command for updates to take effect')}\n"
        )
        print(emojis(s))


def check_img_size(imgsz, s=32, floor=0):
    # Verify image size is a multiple of stride s in each dimension
    if isinstance(imgsz, int):  # integer i.e. img_size=640
        new_size = max(make_divisible(imgsz, int(s)), floor)
    else:  # list i.e. img_size=[640, 480]
        new_size = [max(make_divisible(x, int(s)), floor) for x in imgsz]
    if new_size != imgsz:
        print(
            f"WARNING: --img-size {imgsz} must be multiple of max stride {s}, updating to {new_size}"
        )
    return new_size


def check_suffix(file="yolov5s.pt", suffix=(".pt",), msg=""):
    # Check file(s) for acceptable suffix
    if file and suffix:
        if isinstance(suffix, str):
            suffix = [suffix]
        for f in file if isinstance(file, (list, tuple)) else [file]:
            s = Path(f).suffix.lower()  # file suffix
            if len(s):
                assert s in suffix, f"{msg}{f} acceptable suffix is {suffix}"


def check_yaml(file, suffix=(".yaml", ".yml")):
    # Search/download YAML file (if necessary) and return path, checking suffix
    return check_file(file, suffix)


def check_file(file, suffix=""):
    # Search/download file (if necessary) and return path
    check_suffix(file, suffix)  # optional
    file = str(file)  # convert to str()
    if Path(file).is_file() or file == "":  # exists
        return file
    elif file.startswith(("http:/", "https:/")):  # download
        url = str(Path(file)).replace(":/", "://")  # Pathlib turns :// -> :/
        file = Path(
            urllib.parse.unquote(file).split("?")[0]
        ).name  # '%2F' to '/', split https://url.com/file.txt?auth
        print(f"Downloading {url} to {file}...")
        torch.hub.download_url_to_file(url, file)
        assert (
            Path(file).exists() and Path(file).stat().st_size > 0
        ), f"File download failed: {url}"  # check
        return file
    else:  # search
        files = []
        for d in "data", "models", "utils":  # search directories
            files.extend(
                glob.glob(str(d / "**" / file), recursive=True)
            )  # find file
        assert len(files), f"File not found: {file}"  # assert file was found
        assert (
            len(files) == 1
        ), f"Multiple files match '{file}', specify exact path: {files}"  # assert unique
        return files[0]  # return file

def check_dict(dict1, dict2):
    """This func will check the keys of dict2 in dict1 or not."""
    unexpected = []
    keys = list(dict1.keys())
    for k, _ in dict2.items():
        if k not in keys:
            unexpected.append(k)
    return unexpected



from typing import Any, Dict, Iterable, Iterator, Tuple
from tabulate import tabulate
import inspect


class Registry(Iterable[Tuple[str, Any]]):
    """
    The registry that provides name -> object mapping, to support third-party
    users' custom modules.
    To create a registry (e.g. a backbone registry):
    .. code-block:: python
        BACKBONE_REGISTRY = Registry('BACKBONE')
    To register an object:
    .. code-block:: python
        @BACKBONE_REGISTRY.register()
        class MyBackbone():
            ...
    Or:
    .. code-block:: python
        BACKBONE_REGISTRY.register(MyBackbone)
    """

    def __init__(self, name: str) -> None:
        """
        Args:
            name (str): the name of this registry
        """
        self._name: str = name
        self._obj_map: Dict[str, Any] = {}

    def _do_register(self, name: str, obj: Any) -> None:
        assert (
            name not in self._obj_map
        ), "An object named '{}' was already registered in '{}' registry!".format(
            name, self._name
        )
        self._obj_map[name] = obj

    def register(self, obj: Any = None) -> Any:
        """
        Register the given object under the the name `obj.__name__`.
        Can be used as either a decorator or not. See docstring of this class for usage.
        """
        if obj is None:
            # used as a decorator
            def deco(func_or_class: Any) -> Any:
                name = func_or_class.__name__
                self._do_register(name, func_or_class)
                return func_or_class

            return deco

        # used as a function call
        name = obj.__name__
        self._do_register(name, obj)

    def get(self, name: str) -> Any:
        ret = self._obj_map.get(name)
        if ret is None:
            raise KeyError(
                "No object named '{}' found in '{}' registry!".format(name, self._name)
            )
        return ret

    def __contains__(self, name: str) -> bool:
        return name in self._obj_map

    def __repr__(self) -> str:
        table_headers = ["Names", "Objects"]
        table = tabulate(
            self._obj_map.items(), headers=table_headers, tablefmt="fancy_grid"
        )
        return "Registry of {}:\n".format(self._name) + table

    def __iter__(self) -> Iterator[Tuple[str, Any]]:
        return iter(self._obj_map.items())

    # pyre-fixme[4]: Attribute must be annotated.
    __str__ = __repr__


def build_from_config(cfg, registry: "Registry"):
    if not isinstance(cfg, dict):
        raise TypeError(f"cfg must be a dict, but got {type(cfg)}")
    if "type" not in cfg:
        raise KeyError('`cfg` must contain the key "type"')

    args = cfg.copy()

    obj_type = args.pop("type")
    if isinstance(obj_type, str):
        obj_cls = registry.get(obj_type)
        if obj_cls is None:
            raise KeyError(f"{obj_type} is not in the {registry.name} registry")
    elif inspect.isclass(obj_type) or inspect.isfunction(obj_type):
        obj_cls = obj_type
    else:
        raise TypeError(f"type must be a str or valid type, but got {type(obj_type)}")
    try:
        return obj_cls(**args)
    except Exception as e:
        # Normal TypeError does not print class name.
        raise type(e)(f"{obj_cls.__name__}: {e}")



import cv2

cv2_interp_codes = {
    "nearest": cv2.INTER_NEAREST,
    "bilinear": cv2.INTER_LINEAR,
    "bicubic": cv2.INTER_CUBIC,
    "area": cv2.INTER_AREA,
    "lanczos": cv2.INTER_LANCZOS4,
}


def imrescale(
    img,
    new_wh,
    keep_ratio=True,
    return_scale=False,
    interpolation="bilinear",
):
    """Resize image to a given size.

    Args:
        img (ndarray): The input image.
        new_wh (tuple[int]): Target size (w, h).
        keep_ratio (bool): Whether to keep the aspect ratio.
        return_scale (bool): Whether to return `w_scale` and `h_scale`.
        interpolation (str): Interpolation method, accepted values are
            "nearest", "bilinear", "bicubic", "area", "lanczos" for 'cv2'
            backend, "nearest", "bilinear" for 'pillow' backend.

    Returns:
        tuple | ndarray: (`resized_img`, `w_scale`, `h_scale`) or
        `resized_img`.
    """
    h, w = img.shape[:2]
    nw, nh = new_wh
    if keep_ratio:
        scale_factor = min(nw / w, nh / h)
        img = cv2.resize(
            img,
            (int(w * scale_factor), int(h * scale_factor)),
            interpolation=cv2_interp_codes[interpolation],
        )
    else:
        img = cv2.resize(img, (nw, nh), interpolation=cv2_interp_codes[interpolation])

    if return_scale:
        nh, nw = img.shape[:2]
        w_scale = nw / w
        h_scale = nh / h
        return img, w_scale, h_scale
    return img



from lqcv.bbox.areas import Area, Areas
from lqcv.data import create_reader
from yolov5.core import Yolov5
from tqdm import tqdm
import cv2


def test_one_rect(rect, detector, dataset):
    area = Area(rect, area_type="rect")
    for img, _, _ in tqdm(dataset, total=dataset.frames):
        outputs = detector.inference(img, classes=[0])
        boxes = outputs[0][:, :4].cpu()
        index = area.boxes_in_area(boxes, frame=img, filter="center")
        img = detector.visualize(img, outputs[0][index])
        area.plot(img)
        cv2.imshow("p", cv2.resize(img, (1280, 720)))
        if cv2.waitKey(1) == ord("q"):
            break


def test_multi_rects(rects, detector, dataset):
    areas = Areas(rects, area_type="rect")
    for img, _, _ in tqdm(dataset, total=dataset.frames):
        outputs = detector.inference(img, classes=[0])
        boxes = outputs[0][:, :4].cpu()
        index = areas.boxes_in_areas(boxes, frame=img, filter="center")
        for i in index:
            img = detector.visualize(img, outputs[0][i])
        areas.plot(img)
        cv2.imshow("p", cv2.resize(img, (1280, 720)))
        if cv2.waitKey(1) == ord("q"):
            break


def test_one_polygon(polygon, detector, dataset):
    area = Area(polygon, area_type="polygon")
    for img, _, _ in tqdm(dataset, total=dataset.frames):
        outputs = detector.inference(img, classes=[0])
        boxes = outputs[0][:, :4].cpu()
        index = area.boxes_in_area(boxes, frame=img, filter="lb")
        img = detector.visualize(img, outputs[0][index])
        area.plot(img)
        cv2.imshow("p", cv2.resize(img, (1280, 720)))
        if cv2.waitKey(1) == ord("q"):
            break


def test_multi_polygons(polygons, detector, dataset):
    areas = Areas(polygons, area_type="polygon")
    for img, _, _ in tqdm(dataset, total=dataset.frames):
        outputs = detector.inference(img, classes=[0])
        boxes = outputs[0][:, :4].cpu()
        index = areas.boxes_in_areas(boxes, frame=img, filter="center")
        for i in index:
            img = detector.visualize(img, outputs[0][i])
        areas.plot(img)
        cv2.imshow("p", cv2.resize(img, (1280, 720)))
        if cv2.waitKey(1) == ord("q"):
            break


if __name__ == "__main__":
    rect = [545, 248, 1414, 616]
    polygon = [559, 245, 1035, 242, 1136, 668, 253, 720]

    rects = [[545, 248, 1414, 616], [1325, 216, 1806, 910], [97, 747, 945, 987]]
    polygons = [
        [559, 245, 1035, 242, 1136, 668, 253, 720],
        [1399, 193, 1777, 462, 1536, 715, 1044, 695],
        [115, 714, 512, 720, 821, 883, 502, 1021, 131, 974],
    ]

    detector = Yolov5(weights="./weights/yolov5m.pt", device=0)
    dataset, _ = create_reader(source="./test_videos/test.mp4")

    # test_one_rect(rect, detector, dataset)
    # test_one_polygon(polygon, detector, dataset)

    # test_multi_rects(rects, detector, dataset)
    test_multi_polygons(polygons, detector, dataset)



import numpy as np
import torch
import cv2
from typing import Iterable
from .bbox_utils import Boxes


class Area:
    """
    Boxes in areas or not.
    Args:
        area (List | ndarray): single area,
            if area_type is `rect`, then area is like [x1, y1, x2, y2]
            which x1,y1 is left top and x2,y2 is right bottom.
            if area_type is `polygon`, then area is like [x1, y1, x2, y2, x3, y3,...]
            which each x,y is the vertice of polygon.
        area_type (str): `rect` or `polygon`.
    """

    def __init__(self, area, area_type="rect") -> None:
        self.area = self._check(area, area_type)
        self.area_type = area_type

    def _check(self, area, area_type):
        assert area_type in ["rect", "polygon"]
        assert isinstance(area, Iterable)
        if area_type == "rect":
            assert len(area) == 4
            x1, y1, x2, y2 = area
            assert x2 >= x1 and y2 >= y1
            return area
        else:
            assert len(area) % 2 == 0
            # Polygon vertices need to be connected end to end
            area = area + area[:2]
            area = np.asarray(area)
            area = area.reshape(len(area) // 2, 2)
            return area

    def boxes_in_rect(self, boxes, *args, **kwargs):
        """
        Args:
            boxes (np.ndarray | torch.Tensor): xyxy format, shape: (n, 4) or (4, ).
        Return:
            index (np.ndarray | torch.Tensor): same type and length with boxes, index
                of boxes which are in areas.
        """
        if len(boxes) == 0:
            return False

        left_top = self.area[:2]
        right_bottom = self.area[2:]
        # deal with torch.Tensor
        if isinstance(boxes, torch.Tensor):
            left_top = torch.as_tensor(left_top)
            right_bottom = torch.as_tensor(right_bottom)

        boxes = Boxes(boxes)
        # (m, n, 2)
        vertices = boxes.get_vertice(*args, **kwargs)
        # (m, n)
        count = (vertices > left_top).all(2) * (vertices < right_bottom).all(2)
        # (n, )
        index = count.any(0)
        return index

    def boxes_in_polygon(self, boxes, *args, **kwargs):
        """
        Args:
            boxes (np.ndarray | torch.Tensor): xyxy format, shape: (n, 4) or (4, ).
        Return:
            index (List): index of boxes which are in areas.
        """
        if len(boxes) == 0:
            return False
        boxes = Boxes(boxes)
        # (m, n, 2)
        vertices = boxes.get_vertice(*args, **kwargs)
        # torch.Tensor -> ndarray, cause `_isPointinPolygon` support numpy for now.
        if isinstance(vertices, torch.Tensor):
            vertices = vertices.numpy()
        # (m, n)
        count = [self._isPointinPolygon(vertice, self.area) for vertice in vertices]
        index = np.asarray(count).any(0)
        return index

    def boxes_in_area(self, boxes, *args, **kwargs):
        """
        Args:
            boxes (np.ndarray | torch.Tensor): xyxy format, shape: (n, 4) or (4, ).
        Return:
            index (List): index of boxes which are in areas.
        """
        if self.area_type == "rect":
            return self.boxes_in_rect(boxes, *args, **kwargs)
        else:
            return self.boxes_in_polygon(boxes, *args, **kwargs)

    def plot(self, frame):
        if self.area_type == "rect":
            cv2.rectangle(
                frame, self.area[:2], self.area[2:], (0, 0, 255), 2, cv2.LINE_AA
            )
        else:
            cv2.polylines(
                img=frame,
                pts=[self.area],
                isClosed=True,
                color=(0, 0, 255),
                thickness=3,
            )

    def _isPointinPolygon(self, points, rangelist):  # [[0,0],[1,1],[0,1],[0,0]] [1,0.8]
        """
        :param points: numpy, 坐标xy,shape(num_points, 2)
        :param rangelist: rangelist.shape=(N+1, 2)
        :return: points in rangelist or not
        """
        if len(points) == 0:
            return []

        count = np.zeros_like(points)[:, 0]
        point1 = rangelist[0]
        for i in range(1, len(rangelist)):
            point2 = rangelist[i]
            index = ((point1[1] < points[:, 1]) * (point2[1] >= points[:, 1])) + (
                (point1[1] >= points[:, 1]) * (point2[1] < points[:, 1])
            )

            """
            求线段与射线交点 再和point的lat比较
            先取满足index条件的跑points计算,防止(point2[1] - point1[1])==0，出现nan的情况
            此方法会稍稍慢一些(大约0.0几毫秒)，但比for原本的for循环快多了
            """
            # ind记录索引
            ind = np.nonzero(index)
            points_ = points[index]
            point12lng_part = point2[0] - (point2[1] - points_[:, 1]) * (
                point2[0] - point1[0]
            ) / (point2[1] - point1[1])
            point12lng = np.zeros_like(points)[:, 0]
            point12lng[ind] = point12lng_part
            # print(ind, point12lng)

            # 判断小于<, 表示射线向x轴左边射
            count[(point12lng < points[:, 0]) * point12lng != 0] += 1
            point1 = point2

        return np.asarray(count % 2, dtype=np.bool)


class Areas(Area):
    """
    Args:
        areas (List | ndarray): multi areas,
            if area_type is `rect`, then area is like List[[x1, y1, x2, y2], [...]]
            which x1,y1 is left top and x2,y2 is right bottom.
            if area_type is `polygon`, then area is like List[[x1, y1, x2, y2, x3, y3,...], [...]]
            which each x,y is the vertice of polygon.
        area_type (str): `rect` or `polygon`.
    """

    def __init__(self, areas, area_type="rect") -> None:
        # List[area]
        self.areas = [self._check(area, area_type) for area in areas]
        assert (
            len(self.areas) > 1
        ), "you got only one area, please use single versoin `Area`."
        if area_type == "rect":
            self.areas = np.asarray(self.areas)
        self.area_type = area_type

    def boxes_in_rects(self, boxes, *args, **kwargs):
        """
        Args:
            boxes (np.ndarray | torch.Tensor): xyxy format, shape: (n, 4) or (4, ).
        Return:
            index (np.ndarray | torch.Tensor): same type and length with boxes, index
                of boxes which are in areas.
        """
        if len(boxes) == 0:
            return [False]

        # (na, 2)
        left_top = self.areas[:, :2]
        right_bottom = self.areas[:, 2:]
        # deal with torch.Tensor
        if isinstance(boxes, torch.Tensor):
            left_top = torch.as_tensor(left_top)
            right_bottom = torch.as_tensor(right_bottom)

        boxes = Boxes(boxes)
        # (m, nb, 2)
        vertices = boxes.get_vertice(*args, **kwargs)
        # (na, m, nb)
        count = (vertices[None, :, :, :] > left_top[:, None, None, :]).all(3) * (
            vertices[None, :, :, :] < right_bottom[:, None, None, :]
        ).all(3)
        # (na, nb)
        index = count.any(1)
        return index

    def boxes_in_polygons(self, boxes, *args, **kwargs):
        """
        Args:
            boxes (np.ndarray | torch.Tensor): xyxy format, shape: (n, 4) or (4, ).
        Return:
            index (List): index of boxes which are in areas.
        """
        if len(boxes) == 0:
            return [False]
        boxes = Boxes(boxes)
        # (m, nb, 2)
        vertices = boxes.get_vertice(*args, **kwargs)
        # torch.Tensor -> ndarray, cause `_isPointinPolygon` support numpy for now.
        if isinstance(vertices, torch.Tensor):
            vertices = vertices.numpy()
        # (na, m, nb)
        count = [
            [self._isPointinPolygon(vertice, area) for vertice in vertices]
            for area in self.areas
        ]
        # (na, nb)
        index = np.asarray(count).any(1)
        return index

    def boxes_in_areas(self, boxes, *args, **kwargs):
        """
        Args:
            boxes (np.ndarray | torch.Tensor): xyxy format, shape: (n, 4) or (4, ).
        Return:
            index (List): index of boxes which are in areas.
        """
        if self.area_type == "rect":
            return self.boxes_in_rects(boxes, *args, **kwargs)
        else:
            return self.boxes_in_polygons(boxes, *args, **kwargs)

    def plot(self, frame):
        for area in self.areas:
            if self.area_type == "rect":
                cv2.rectangle(frame, area[:2], area[2:], (0, 0, 255), 2, cv2.LINE_AA)
            else:
                cv2.polylines(
                    img=frame,
                    pts=[area],
                    isClosed=True,
                    color=(0, 0, 255),
                    thickness=3,
                )
